@@ -9,6 +9,7 @@
 #include <stdexcept>
 #include <stdint.h>
 #include <stdio.h>
+#include <vector>
 
 namespace rapidfuzz {
 namespace detail {
@@ -45,37 +46,22 @@ struct BitMatrix {
 
     using value_type = T;
 
-    BitMatrix() : m_rows(0), m_cols(0), m_matrix(nullptr), m_start_offset(0), m_offset_per_row(0)
+    BitMatrix() : m_rows(0), m_cols(0), m_matrix(nullptr)
     {}
 
-    BitMatrix(size_t rows, size_t cols, T val, ptrdiff_t start_offset = 0, ptrdiff_t offset_per_row = 0)
-        : m_rows(rows),
-          m_cols(cols),
-          m_matrix(nullptr),
-          m_start_offset(start_offset),
-          m_offset_per_row(offset_per_row)
+    BitMatrix(size_t rows, size_t cols, T val) : m_rows(rows), m_cols(cols), m_matrix(nullptr)
     {
         if (m_rows && m_cols) m_matrix = new T[m_rows * m_cols];
         std::fill_n(m_matrix, m_rows * m_cols, val);
     }
 
-    BitMatrix(const BitMatrix& other)
-        : m_rows(other.m_rows),
-          m_cols(other.m_cols),
-          m_matrix(nullptr),
-          m_start_offset(other.m_start_offset),
-          m_offset_per_row(other.m_offset_per_row)
+    BitMatrix(const BitMatrix& other) : m_rows(other.m_rows), m_cols(other.m_cols), m_matrix(nullptr)
     {
         if (m_rows && m_cols) m_matrix = new T[m_rows * m_cols];
         std::copy(other.m_matrix, other.m_matrix + m_rows * m_cols, m_matrix);
     }
 
-    BitMatrix(BitMatrix&& other) noexcept
-        : m_rows(0),
-          m_cols(0),
-          m_matrix(nullptr),
-          m_start_offset(other.m_start_offset),
-          m_offset_per_row(other.m_offset_per_row)
+    BitMatrix(BitMatrix&& other) noexcept : m_rows(0), m_cols(0), m_matrix(nullptr)
     {
         other.swap(*this);
     }
@@ -99,8 +85,6 @@ struct BitMatrix {
         swap(m_rows, rhs.m_rows);
         swap(m_cols, rhs.m_cols);
         swap(m_matrix, rhs.m_matrix);
-        swap(m_start_offset, rhs.m_start_offset);
-        swap(m_offset_per_row, rhs.m_offset_per_row);
     }
 
     ~BitMatrix()
@@ -108,28 +92,6 @@ struct BitMatrix {
         delete[] m_matrix;
     }
 
-    bool test_bit(size_t row, size_t col, bool default_ = false) const noexcept
-    {
-        ptrdiff_t offset = m_start_offset + static_cast<ptrdiff_t>(row) * m_offset_per_row;
-
-        if (offset < 0) {
-            col += static_cast<size_t>(-offset);
-        }
-        else if (col >= static_cast<size_t>(offset)) {
-            col -= static_cast<size_t>(offset);
-        }
-        /* bit on the left of the band */
-        else {
-            return default_;
-        }
-
-        size_t word_size = sizeof(value_type) * 8;
-        size_t col_word = col / word_size;
-        uint64_t col_mask = value_type(1) << (col % word_size);
-
-        return bool(m_matrix[row * m_cols + col_word] & col_mask);
-    }
-
     BitMatrixView<value_type, false> operator[](size_t row) noexcept
     {
         assert(row < m_rows);
@@ -156,8 +118,86 @@ private:
     size_t m_rows;
     size_t m_cols;
     T* m_matrix;
-    ptrdiff_t m_start_offset;
-    ptrdiff_t m_offset_per_row;
+};
+
+template <typename T>
+struct ShiftedBitMatrix {
+    using value_type = T;
+
+    ShiftedBitMatrix()
+    {}
+
+    ShiftedBitMatrix(size_t rows, size_t cols, T val) : m_matrix(rows, cols, val), m_offsets(rows)
+    {}
+
+    ShiftedBitMatrix(const ShiftedBitMatrix& other) : m_matrix(other.m_matrix), m_offsets(other.m_offsets)
+    {}
+
+    ShiftedBitMatrix(ShiftedBitMatrix&& other) noexcept
+    {
+        other.swap(*this);
+    }
+
+    ShiftedBitMatrix& operator=(ShiftedBitMatrix&& other) noexcept
+    {
+        other.swap(*this);
+        return *this;
+    }
+
+    ShiftedBitMatrix& operator=(const ShiftedBitMatrix& other)
+    {
+        ShiftedBitMatrix temp = other;
+        temp.swap(*this);
+        return *this;
+    }
+
+    void swap(ShiftedBitMatrix& rhs) noexcept
+    {
+        using std::swap;
+        swap(m_matrix, rhs.m_matrix);
+        swap(m_offsets, rhs.m_offsets);
+    }
+
+    bool test_bit(size_t row, size_t col, bool default_ = false) const noexcept
+    {
+        ptrdiff_t offset = static_cast<ptrdiff_t>(m_offsets[row]);
+
+        if (offset < 0) {
+            col += static_cast<size_t>(-offset);
+        }
+        else if (col >= static_cast<size_t>(offset)) {
+            col -= static_cast<size_t>(offset);
+        }
+        /* bit on the left of the band */
+        else {
+            return default_;
+        }
+
+        size_t word_size = sizeof(value_type) * 8;
+        size_t col_word = col / word_size;
+        uint64_t col_mask = value_type(1) << (col % word_size);
+
+        return bool(m_matrix[row][col_word] & col_mask);
+    }
+
+    auto operator[](size_t row) noexcept
+    {
+        return m_matrix[row];
+    }
+
+    auto operator[](size_t row) const noexcept
+    {
+        return m_matrix[row];
+    }
+
+    void set_offset(size_t row, ptrdiff_t offset)
+    {
+        m_offsets[row] = offset;
+    }
+
+private:
+    BitMatrix<value_type> m_matrix;
+    std::vector<ptrdiff_t> m_offsets;
 };
 
 } // namespace detail
